from japick.lines import Line
from japick.paragraph import Paragraph, Pos, parse_paragraph
from japick.syntax import NULL_SYMBOL


class TestParagraph:
    def test_body(self):
        target = Paragraph(
            Pos(0, 0),
            0,
            [
                "  これはテスト。",
                f"  次の {NULL_SYMBOL * 2}テスト{NULL_SYMBOL * 2}。",
            ],
        )
        assert (
            target.body
            == """これはテスト。
次の テスト。"""
        )

    def test_body_memoize(self):
        target = Paragraph(0, 0, [], _body="Memoized")
        assert target.body == "Memoized"

    def test_as_original_pos(self):
        target = Paragraph(
            Pos(0, 0),
            0,
            [
                "  これはテスト。",
                f"  次の {NULL_SYMBOL * 2}テスト{NULL_SYMBOL * 2}。",
            ],
        )
        # これはテスト。
        # 次の テ[HERE]スト。
        actual = target.as_original_pos(12)
        assert actual == Pos(1, 8)

    def test_as_original_pos_greedy_lazy(self):
        target = Paragraph(
            Pos(0, 0),
            0,
            [f"次の {NULL_SYMBOL * 2}テスト{NULL_SYMBOL * 2}。"],
        )
        # 次の テスト[HERE]。
        actual = target.as_original_pos(3, lazy=False)
        assert actual == Pos(0, 5)
        actual = target.as_original_pos(6, lazy=False)
        assert actual == Pos(0, 10)
        actual = target.as_original_pos(3, lazy=True)
        assert actual == Pos(0, 3)
        actual = target.as_original_pos(6, lazy=True)
        assert actual == Pos(0, 8)

    def test_as_original_index(self):
        target = Paragraph(
            Pos(5, 0),
            200,
            [
                "    これはテスト。",
                "      これは2行目",
            ],
        )
        actual = target.as_original_index(11)
        assert actual == 221

    def test_as_original_index_lazy(self):
        target = Paragraph(Pos(0, 0), 0, [f"これは{NULL_SYMBOL}テスト。"])
        assert target.as_original_index(3, lazy=False) == 4
        assert target.as_original_index(3, lazy=True) == 3


def test_parse_paragraph_list():
    actual = parse_paragraph(
        [
            Line("文"),
            Line("リスト", is_list=True),
            Line("リスト2", is_list=True),
        ]
    )
    assert list(actual) == [
        Paragraph(Pos(0, 0), 0, ["文"]),
        Paragraph(Pos(1, 0), 2, ["リスト"], is_list=True),
        Paragraph(Pos(2, 0), 6, ["リスト2"], is_list=True),
    ]


def test_parse_paragraph_heading():
    actual = list(
        parse_paragraph(
            [
                Line("見出し", is_head=True),
                Line("文"),
                Line("文2"),
            ]
        )
    )
    assert actual == [
        Paragraph(Pos(0, 0), 0, ["見出し"], is_head=True),
        Paragraph(Pos(1, 0), 4, ["文", "文2"]),
    ]


def test_parse_paragraph_multiple():
    actual = parse_paragraph(
        [
            Line("Test", is_ignore=True),
            Line("文1"),
            Line("文2"),
            Line("", is_ignore=True),
            Line("文3"),
            Line("文4"),
        ]
    )
    assert list(actual) == [
        Paragraph(Pos(1, 0), 5, ["文1", "文2"]),
        Paragraph(Pos(4, 0), 12, ["文3", "文4"]),
    ]



import re
from typing import Callable

from .syntax import (
    CODE_REGEX,
    CODE_TAG_REGEX,
    HEAD_REGEX,
    HTML_REGEX,
    INLINE_RETURN_SYMBOL,
    LINK_REGEX,
    LIST_REGEX,
    MASK_SYMBOL,
    NAROU_POINTS,
    NAROU_RUBY,
    NULL_SYMBOL,
    QUOTE_REGEX,
    SYMBOL_REGEX,
    URL_REGEX,
)

NOT_RET_REGEX = re.compile(r"[^\n]")


def fill(symbol: str) -> Callable:
    def _runner(matched) -> str:
        return symbol * len(matched[0])

    return _runner


def clean(line):
    """
    * Remove URL
    * Remove Markdown symbols and fill by spaces.
        * List * - 1. * [ ] ...
        * Quote > ...
        * Heading # ...
        * Link and Image [...]()
        * Bold * _
        * Strikethrough ~
    * Mask texts inside Codes `...`
    """
    line.body = LIST_REGEX.sub(fill(" "), line.body)
    line.body = QUOTE_REGEX.sub(fill(" "), line.body)
    line.body = HEAD_REGEX.sub(fill(" "), line.body)
    line.body = LINK_REGEX.sub(
        lambda m: NULL_SYMBOL * len(m[1]) + m[2] + NULL_SYMBOL * len(m[3]), line.body
    )
    line.body = HTML_REGEX.sub(fill(NULL_SYMBOL), line.body)
    line.body = URL_REGEX.sub(fill(NULL_SYMBOL), line.body)
    line.body = SYMBOL_REGEX.sub(fill(NULL_SYMBOL), line.body)
    line.body = CODE_REGEX.sub(fill(MASK_SYMBOL), line.body)
    line.body = NAROU_RUBY.sub(
        lambda m: f" {m['text']} {' ' * len(m['ruby'])} ",
        line.body,
    )
    line.body = NAROU_POINTS.sub(
        lambda m: f"  {m['text']}  ",
        line.body,
    )
    return line


def fill_null_symbols(b):
    b = b.replace(INLINE_RETURN_SYMBOL, "o")
    b = NOT_RET_REGEX.sub(NULL_SYMBOL, b)
    b = b.replace("\n", INLINE_RETURN_SYMBOL)
    return b


def unescape_html(body: str):
    body = (
        body.replace("&lt;", "<")
        .replace("&gt;", ">")
        .replace("&quot;", '"')
        .replace("&apos;", "'")
        .replace("&nbsp;", " ")
        .replace("&amp;", "&")
    )
    return body


def clean_html(body: str):
    body = CODE_TAG_REGEX.sub(
        lambda m: fill_null_symbols(m[1])
        + NOT_RET_REGEX.sub(MASK_SYMBOL, m[2])
        + fill_null_symbols(m[3]),
        body,
    )
    body = HTML_REGEX.sub(lambda m: fill_null_symbols(m[0]), body)
    return unescape_html(body)


def clean_lines(lines):
    for line in lines:
        if not line.is_ignore:
            clean(line)
        yield line


def clean_cr_lf(body: str) -> str:
    return body.replace("\r\n", " \n").replace("\r", "\n")



from japick.cleaner import clean_lines
from japick.lines import Line, parse_lines
from japick.syntax import MASK_SYMBOL, NULL_SYMBOL


def call_target(body):
    lines = parse_lines(body)
    return list(clean_lines(lines))


def test_clean_list():
    actual = call_target(
        """* リスト1
* リスト2
    - リスト3
    - [ ] リスト4"""
    )
    assert actual == [
        Line("  リスト1", is_list=True),
        Line("  リスト2", is_list=True),
        Line("      リスト3", is_list=True),
        Line("          リスト4", is_list=True),
    ]


def test_clean_quote():
    actual = call_target(
        """> テスト
> テスト2"""
    )
    assert actual == [
        Line("  テスト"),
        Line("  テスト2"),
    ]


def test_clean_head():
    actual = call_target(
        """# 見出し1
## 見出し2
"""
    )
    assert actual == [
        Line("  見出し1", is_head=True),
        Line("   見出し2", is_head=True),
    ]


def test_clean_link():
    actual = call_target(
        """これは ![テスト](https://foobar.com/) です。
実際には[リンク](https://shodo.ink/)となります。
難しいことはありません [1]。"""
    )
    assert actual == [
        Line(f"これは {NULL_SYMBOL * 2}テスト{NULL_SYMBOL * 22} です。"),
        Line(f"実際には{NULL_SYMBOL}リンク{NULL_SYMBOL * 21}となります。"),
        Line("難しいことはありません [1]。"),
    ]


def test_clean_html():
    actual = call_target(
        """<section>
    テスト</section>
これは <a href="#">リンク</a> のテスト。"""
    )
    assert actual == [
        Line("<section>", is_ignore=True),
        Line(f"    テスト{NULL_SYMBOL * 10}"),
        Line(f"これは {NULL_SYMBOL * 12}リンク{NULL_SYMBOL * 4} のテスト。"),
    ]


def test_clean_html_close_tag_text():
    actual = call_target("""<a data-example=">">テスト</a>""")
    assert actual == [
        Line(f"{NULL_SYMBOL * 20}テスト{NULL_SYMBOL * 4}"),
    ]


def test_clean_url():
    actual = call_target(
        """次は https://shodo.ink/ でお会いしましょう。
URLに https://foobar.com/日本語/ が入っている場合。
括弧書きは（https://foobar.com/）機能します。
"""
    )
    assert actual == [
        Line(f"次は {NULL_SYMBOL * 18} でお会いしましょう。"),
        Line(f"URLに {NULL_SYMBOL * 19}日本語/ が入っている場合。"),
        Line(f"括弧書きは（{NULL_SYMBOL * 19}）機能します。"),
    ]


def test_clean_symbol():
    actual = call_target(
        """**太字** の場合。
__別の太字__ もある。
間に test_foo と入る。
他に,論文っぽい読点は許容する (半角カッコも) .
"""
    )
    assert actual == [
        Line(f"{NULL_SYMBOL * 2}太字{NULL_SYMBOL * 2} の場合。"),
        Line(f"{NULL_SYMBOL * 2}別の太字{NULL_SYMBOL * 2} もある。"),
        Line("間に test_foo と入る。"),
        Line("他に,論文っぽい読点は許容する (半角カッコも) ."),
    ]


def test_mask_code():
    actual = call_target(
        """コード表記は `mask` される。
理由は単純で、コードを形態素解析 `しない` ため。"""
    )
    assert actual == [
        Line(f"コード表記は {MASK_SYMBOL * 6} される。"),
        Line(f"理由は単純で、コードを形態素解析 {MASK_SYMBOL * 5} ため。"),
    ]


def test_it():
    actual = call_target(
        """
# 見出し1
## 見出し2
### 見出し3

- リスト1
    - ネスト リスト1_1
        - ネスト リスト1_1_1
    - ネスト リスト1_2
- リスト2

1. 番号付きリスト1
    1. 番号付きリスト1_1
1. 番号付きリスト2

> ご連絡いただいた
> お世話になります！
>> お世話になります。

    # コメント
    class Test
        pass

インストールは `pip install japick` です。
**太字** の場合。
___別の太字___ も
[Googleさん](https://www.google.co.jp/) 。
~~取り消し~~。[内部リンク][google]。
"""
    )
    assert actual == [
        Line("", is_ignore=True),
        Line("  見出し1", is_head=True),
        Line("   見出し2", is_head=True),
        Line("    見出し3", is_head=True),
        Line("", is_ignore=True),
        Line("  リスト1", is_list=True),
        Line("      ネスト リスト1_1", is_list=True),
        Line("          ネスト リスト1_1_1", is_list=True),
        Line("      ネスト リスト1_2", is_list=True),
        Line("  リスト2", is_list=True),
        Line("", is_ignore=True),
        Line("   番号付きリスト1", is_list=True),
        Line("       番号付きリスト1_1", is_list=True),
        Line("   番号付きリスト2", is_list=True),
        Line("", is_ignore=True),
        Line("  ご連絡いただいた"),
        Line("  お世話になります！"),
        Line("   お世話になります。"),
        Line("", is_ignore=True),
        Line("    # コメント"),
        Line("    class Test", is_ignore=True),
        Line("        pass", is_ignore=True),
        Line("", is_ignore=True),
        Line(f"インストールは {MASK_SYMBOL * 20} です。"),
        Line(f"{NULL_SYMBOL * 2}太字{NULL_SYMBOL * 2} の場合。"),
        Line(f"{NULL_SYMBOL * 3}別の太字{NULL_SYMBOL * 3} も"),
        Line(f"{NULL_SYMBOL}Googleさん{NULL_SYMBOL * 28} 。"),
        Line(f"{NULL_SYMBOL * 2}取り消し{NULL_SYMBOL * 2}。[内部リンク][google]。"),
    ]


def test_clean_narou():
    actual = call_target(
        """\
これはテストです。
《《何を言っているのかわからねーと思うが》》。
｜伸縮自在の愛《バンジーガム》
|薄っぺらな嘘《ドッキリテクスチャー》
|これは|テスト|"""
    )
    assert actual == [
        Line("これはテストです。", is_list=False, is_head=False),
        Line("  何を言っているのかわからねーと思うが  。", is_list=False, is_head=False),
        Line(" 伸縮自在の愛        ", is_list=False, is_head=False),
        Line(" 薄っぺらな嘘            ", is_list=False, is_head=False),
        Line("|これは|テスト|", is_list=False, is_head=False),
    ]



from dataclasses import dataclass
from typing import Iterable

from .syntax import FENCE_END, FENCE_START, HEAD_REGEX, JA_REGEX, LIST_REGEX


class IgnoreBlock:
    def __init__(self):
        self.inside = False
        self.start_regex = FENCE_START
        self.end_regex = FENCE_END

    def should_ignore(self, line: str) -> bool:
        if self.inside:
            if self.end_regex.search(line):
                self.inside = False
            return True
        else:
            if self.start_regex.search(line):
                self.inside = True
                return True
            else:
                return False


@dataclass
class Line:
    """Annotation for each body lines.
    japick will clean all of syntax of original text,
    so before the process, We need to annotate information.
    """

    body: str
    is_head: bool = False
    is_list: bool = False
    is_ignore: bool = False

    def __eq__(self, other) -> bool:
        if other is None:
            return False
        return (
            self.body == other.body
            and self.is_head is other.is_head
            and self.is_list is other.is_list
            and self.is_ignore is other.is_ignore
        )


def parse_lines(body: str) -> Iterable[Line]:
    lines = body.splitlines()
    is_ignore = IgnoreBlock()

    for line in lines:
        if is_ignore.should_ignore(line):
            yield Line(line, is_ignore=True)
        elif JA_REGEX.search(line) is None:
            yield Line(line, is_ignore=True)
        elif HEAD_REGEX.search(line):
            yield Line(line, is_head=True)
        elif LIST_REGEX.search(line):
            yield Line(line, is_list=True)
        else:
            yield Line(line)



from japick.lines import Line, parse_lines


class TestParseLines:
    def test_fenced(self):
        actual = parse_lines(
            """```python
# これはテストです。
import this
```"""
        )
        assert list(actual) == [
            Line("```python", is_ignore=True),
            Line("# これはテストです。", is_ignore=True),
            Line("import this", is_ignore=True),
            Line("```", is_ignore=True),
        ]

    def test_japanese(self):
        actual = parse_lines(
            """This is test
日本語の文章を取得します。
123-4567
\u2E80
</section>"""
        )
        assert list(actual) == [
            Line("This is test", is_ignore=True),
            Line("日本語の文章を取得します。"),
            Line("123-4567", is_ignore=True),
            Line("\u2E80"),
            Line("</section>", is_ignore=True),
        ]

    def test_head(self):
        actual = parse_lines(
            """# テストです
## テストの見出し2
###許容する見出し
# This should be ignored."""
        )
        assert list(actual) == [
            Line("# テストです", is_head=True),
            Line("## テストの見出し2", is_head=True),
            Line("###許容する見出し", is_head=True),
            Line("# This should be ignored.", is_ignore=True),
        ]

    def test_list(self):
        actual = parse_lines(
            """次に示します。
* リスト1
* リスト2
- リスト3
- リスト4
    - リスト5
    - [ ] リスト6
    - [x]リスト7
1. リスト8
    1. リスト9
    2. リスト10"""
        )
        assert list(actual) == [
            Line("次に示します。"),
            Line("* リスト1", is_list=True),
            Line("* リスト2", is_list=True),
            Line("- リスト3", is_list=True),
            Line("- リスト4", is_list=True),
            Line("    - リスト5", is_list=True),
            Line("    - [ ] リスト6", is_list=True),
            Line("    - [x]リスト7", is_list=True),
            Line("1. リスト8", is_list=True),
            Line("    1. リスト9", is_list=True),
            Line("    2. リスト10", is_list=True),
        ]



"""
Text Parser:
1. Split body line by line
2. Detect Japanese inside each lines
3. Detect line types
    * Header
    * List
    * General
4. Clean up Japanese lines
    * Remove URLs
    * Remove ASCII symbols
5. Detect Paragraphs and wrap by models

HTML Parser:
1. Parse HTML tags
2. Detect Japanese inside tags
"""

from typing import Iterable

from .cleaner import clean_cr_lf, clean_html, clean_lines
from .lines import parse_lines
from .paragraph import Paragraph, Pos, parse_paragraph
from .parser import html_parser
from .syntax import JA_REGEX


def parse(body: str) -> Iterable[Paragraph]:
    body = clean_cr_lf(body)
    lines = parse_lines(body)
    lines = clean_lines(lines)
    return parse_paragraph(lines)


def parse_html(html: str) -> Iterable[Paragraph]:
    html = clean_cr_lf(html)
    tags = html_parser(html)
    for tag in tags:
        lines = clean_html(tag["body"]).splitlines()
        if all(JA_REGEX.search(line) is None for line in lines):
            continue

        p = Paragraph(
            Pos(tag["pos"][0], tag["pos"][1]),
            tag["start"],
            lines,
            is_head=tag["tag"].startswith("h"),
            is_list=tag["tag"] == "li",
        )
        if p.body.strip():
            yield p



import re

NULL_SYMBOL = "~"
INLINE_RETURN_SYMBOL = (
    "ṏ"  # Return symbol inside inline tags (Representing both \n and NULL_SYMBOL)
)
MASK_SYMBOL = "_"

JA_REGEX = re.compile(
    "["
    "\uFF61-\uFF9F"  # HALFWIDTH KATAKANA
    "\u3041-\u309F"  # HIRAGANA
    "\u30A1-\u30FF"  # KATAKANA
    "\u2E80-\u2FDF\u3005-\u3007\u3400-\u4DBF\u4E00-\u9FFF\uF900-\uFAFF"  # KANJI
    "\u2105-\u273D\u3004-\u33FE\uFD3E-\uFFE4"  # JAPANES SPECIAL SYMBOLS
    "\U00020000-\U0002EBEF"  # KANJI
    "]+"
)

HEAD_REGEX = re.compile("^#+")
LIST_REGEX = re.compile(r"^>?\s*([*-] |\d+\. |[･・])(\[[x ]?] )?")
QUOTE_REGEX = re.compile(r"^>+\s*")
HTML_REGEX = re.compile(r'<("[^"]*"|\'[^\']*\'|[^\'">])*>')

URL_REGEX = re.compile(r"<?(http|https|ftp)://[-a-zA-Z0-9@:%_+.~#?&/=]+>?")
LINK_REGEX = re.compile(r"(!?\[)(.*?)(]\(.*?\))")
CODE_REGEX = re.compile(r"`.+?`")
CODE_TAG_REGEX = re.compile(r"(<code\b[^>]*>)(.*?)(</\s*code\s*>)", re.DOTALL | re.IGNORECASE)
SYMBOL_REGEX = re.compile(r"(\*|___|__|~)")  # ASCII Symbols to clean

FENCE_START = re.compile("^```.*$")
FENCE_END = re.compile("^```$")

NAROU_RUBY = re.compile(r"[\|｜](?P<text>.*?)《(?P<ruby>.*?)》")
NAROU_POINTS = re.compile(r"《《(?P<text>.*?)》》")



import re
from dataclasses import dataclass
from typing import Iterable, List, Optional

from .lines import Line
from .syntax import INLINE_RETURN_SYMBOL, NULL_SYMBOL

NULL_MATCH_REGEX = re.compile(f"({re.escape(NULL_SYMBOL)}+)")


@dataclass
class Pos:
    """Line number and column position.
    These numbers starts from Zero.
    """

    line: int
    ch: int

    def __eq__(self, other) -> bool:
        if isinstance(other, tuple):
            return self.line == other[0] and self.ch == other[1]
        else:
            return self.line == other.line and self.ch == other.ch

    def __repr__(self) -> str:
        return f"Pos({self.line}, {self.ch})"


@dataclass
class Paragraph:
    pos: Pos
    index: int
    lines: List[str]
    is_head: bool = False
    is_list: bool = False
    _body: Optional[str] = None

    def __eq__(self, other):
        return (
            self.pos == other.pos
            and self.index == other.index
            and self.lines == other.lines
            and self.is_head is other.is_head
            and self.is_list is other.is_list
        )

    @property
    def body(self):
        if self._body is not None:
            return self._body
        self._body = "\n".join(
            line.strip().replace(INLINE_RETURN_SYMBOL, "").replace(NULL_SYMBOL, "")
            for line in self.lines
        )
        return self._body

    def as_original_pos(self, index, lazy=False) -> Pos:
        """Convert index number of self.body into original text position."""
        # Convert index to cleaned body position.
        # .splitlines("") will be [], so use split("\n")
        body_lines = self.body[:index].split("\n")
        pos = Pos(
            len(body_lines) - 1,
            len(body_lines[-1]),
        )

        # Convert column number to original (self.lines) position.
        line = self.lines[pos.line]
        left_margin = len(line) - len(line.lstrip())

        # Count null symbols before the body position.
        chunks = NULL_MATCH_REGEX.split(
            line[left_margin:].replace(INLINE_RETURN_SYMBOL, NULL_SYMBOL)
        )
        body_chunks = chunks[::2]
        null_chunks = chunks[1::2]
        current = 0
        for i, chunk in enumerate(body_chunks):
            current += len(chunk)
            if (not lazy and pos.ch < current) or (lazy and pos.ch <= current):
                break
        num_null_symbols = sum(len(c) for c in null_chunks[:i])

        # Adjust column number to original position
        pos.ch += left_margin + num_null_symbols

        # Add inline breaks before this line.
        pos.line += sum(li.count(INLINE_RETURN_SYMBOL) for li in self.lines[: pos.line])

        # Add inline breaks on this line and adjust column number.
        inline_break_lines = line[: pos.ch].split(INLINE_RETURN_SYMBOL)
        if len(inline_break_lines) > 1:
            pos.line += len(inline_break_lines) - 1
            pos.ch = len(inline_break_lines[-1])

        # Add relative position to this paragraph to make it absolute of original text.
        pos.ch += self.pos.ch if pos.line == 0 else 0
        pos.line += self.pos.line

        return pos

    def as_original_index(self, index: int, lazy=False) -> int:
        """Convert index number of self.body into original text index."""
        pos = self.as_original_pos(index, lazy=lazy)
        num_lines = pos.line - self.pos.line
        lines = [li for line in self.lines for li in line.split(INLINE_RETURN_SYMBOL)]
        before_index = sum(len(line) for line in lines[:num_lines]) + num_lines
        return self.index + before_index + pos.ch - (self.pos.ch if num_lines == 0 else 0)

    def __str__(self) -> str:
        return self.body

    def __repr__(self) -> str:
        body = self.body
        if self.is_head:
            body = "# " + body
        elif self.is_list:
            body = "* " + body
        return f"Paragraph({self.pos.line},{self.pos.ch}. ({self.index}) {body})"

    def append(self, line: str):
        self.lines.append(line)

    def __bool__(self) -> bool:
        return bool(self.lines)


def parse_paragraph(lines: Iterable[Line]) -> Iterable[Paragraph]:
    """
    Combine multiple lines into one Paragraph.
    Each paragraphs has chunks.
    """
    paragraph = Paragraph(Pos(0, 0), 0, [])
    index = 0
    for i, line in enumerate(lines):
        next_index = index + len(line.body) + 1

        if line.is_ignore or line.is_head or line.is_list:
            if paragraph:
                yield paragraph
            paragraph = Paragraph(Pos(i + 1, 0), next_index, [])

        if line.is_head or line.is_list:
            # Yield my self as 1 line paragraph.
            yield Paragraph(
                Pos(i, 0),
                index,
                [line.body],
                is_head=line.is_head,
                is_list=line.is_list,
            )
        elif not line.is_ignore:
            paragraph.append(line.body)

        index = next_index

    if paragraph:
        yield paragraph



from japick.main import parse, parse_html
from japick.syntax import MASK_SYMBOL


def test_it():
    body = """# これはテストです
こんにちは。
これは文章です。

* リスト
* リスト2と `code` もあります

<section>
  ここは別のパラグラフです。
  Shodo（https://shodo.ink/）とこの次の文字、Xのインデックスを取りましょう。
</section>"""
    paragraphs = list(parse(body))
    assert len(paragraphs) == 5
    assert paragraphs[0].is_head
    assert paragraphs[0].body == "これはテストです"
    assert paragraphs[1].body == "こんにちは。\nこれは文章です。"
    assert paragraphs[2].is_list
    assert paragraphs[2].body == "リスト"
    assert paragraphs[3].is_list
    assert paragraphs[3].body == f"リスト2と {MASK_SYMBOL * 6} もあります"
    assert paragraphs[3].as_original_pos(15) == (5, 17)
    assert paragraphs[3].as_original_index(15) == 51
    assert body[51] == "り"
    assert (
        paragraphs[4].body
        == """\
ここは別のパラグラフです。
Shodo（）とこの次の文字、Xのインデックスを取りましょう。"""
    )
    assert paragraphs[4].as_original_pos(29) == (9, 35)
    assert paragraphs[4].as_original_index(29) == 117
    assert body[117] == "X"


def test_parse_windows_crlf():
    body = "テスト1\r\nテスト2\r\n\r\nテスト3"
    paragraphs = list(parse(body))
    assert len(paragraphs) == 2
    assert paragraphs[0].body == "テスト1\nテスト2"
    assert paragraphs[0].as_original_pos(5) == (1, 0)
    assert paragraphs[0].as_original_index(5) == 6
    assert paragraphs[1].body == "テスト3"


def test_html():
    html = """<h1>これはテストです</h1>
<p>こんにちは。</p>
<ul>
  <li>リスト</li>
  <li>リスト2と <code class="foo">code</code> もあります</li>
</ul>
  <p>
    ここは別のパラ
    グラフです。
  </p>
"""
    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 5
    assert paragraphs[0].is_head
    assert paragraphs[0].body == "これはテストです"
    assert paragraphs[0].as_original_index(3) == 7

    assert not paragraphs[1].is_list
    assert not paragraphs[1].is_head
    assert paragraphs[1].body == "こんにちは。"

    assert paragraphs[2].is_list
    assert paragraphs[2].body == "リスト"

    assert paragraphs[3].is_list
    assert paragraphs[3].body == f"リスト2と {MASK_SYMBOL * 4} もあります"
    assert paragraphs[3].as_original_index(6) == 82
    assert paragraphs[3].as_original_pos(6) == (4, 30)
    assert paragraphs[3].as_original_pos(10, lazy=True) == (4, 34)
    assert paragraphs[3].as_original_pos(10) == (4, 41)

    assert paragraphs[4].body == "\nここは別のパラ\nグラフです。\n"
    assert paragraphs[4].as_original_index(10) == 134
    assert paragraphs[4].as_original_pos(10) == (8, 5)


def test_html_inline_break():
    html = """<p>
  これで<a
  >見れる</a>
</p>"""
    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 1
    assert paragraphs[0].body == "\nこれで見れる\n"
    assert paragraphs[0].as_original_index(4) == 15
    assert paragraphs[0].as_original_pos(4) == (2, 3)
    assert paragraphs[0].as_original_index(4, lazy=True) == 9
    assert paragraphs[0].as_original_pos(4, lazy=True) == (1, 5)


def test_html_inline_breaks_before_the_line():
    html = """
<p class="content">
  <a
  href="http://www.yahoo.co.jp"
  >YAHOO!</a>の検索はこちらから。
  Djangoのインストールは見れる。
</p>
"""

    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 1
    assert paragraphs[0].body == "\nYAHOO!の検索はこちらから。\nDjangoのインストールは見れる。\n"
    assert paragraphs[0].as_original_index(32) == 98


def test_html_inline_breaks_at_the_first():
    html = """  <p><a
    href="http://www.yahoo.co.jp"
    >Djangoのインストールは見れる。
  </a></p>
"""

    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 1
    assert paragraphs[0].body == "Djangoのインストールは見れる。\n"
    assert paragraphs[0].as_original_index(14) == 61


def test_html_inline_breaks_empty_breaks():
    html = """<p><a
    >
    Djangoのインストールは見れる。
  </a></p>
"""

    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 1
    assert paragraphs[0].body == "\nDjangoのインストールは見れる。\n"
    assert paragraphs[0].as_original_index(15) == 30
    assert paragraphs[0].as_original_pos(15) == (2, 18)


def test_html_code():
    html = """<p>
  これは<code
  >KeyWord
   <span/></code>です。
</p>"""
    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 1
    assert paragraphs[0].body == "\nこれは_______\n__________です。\n"
    assert paragraphs[0].as_original_index(4) == 18
    assert paragraphs[0].as_original_pos(4) == (2, 3)
    assert paragraphs[0].as_original_index(4, lazy=True) == 9
    assert paragraphs[0].as_original_pos(4, lazy=True) == (1, 5)


def test_html_upper():
    html = """<P>
  これで見れる
</P>"""
    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 1
    assert paragraphs[0].body == "\nこれで見れる\n"


def test_html_pre_code():
    html = """<pre><code>
  # これはテスト
</code></pre>"""
    assert len(list(parse_html(html))) == 0


def test_windows_crlf():
    html = "<p>テスト1</p>\r\n<p>テスト2</p>\r\n<p>テスト3</p>\r\n<p>テスト4</p>"
    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 4
    assert paragraphs[0].body == "テスト1"
    assert paragraphs[1].body == "テスト2"
    assert paragraphs[1].as_original_pos(3) == (1, 6)
    assert paragraphs[1].as_original_index(3) == 19
    assert paragraphs[2].body == "テスト3"
    assert paragraphs[3].body == "テスト4"
    assert paragraphs[3].as_original_pos(3) == (3, 6)
    assert paragraphs[3].as_original_index(3) == 45


def test_windows_crlf_inside_tag():
    html = "<p>\r\nテスト\r\n1\r\n</p>"
    paragraphs = list(parse_html(html))
    assert len(paragraphs) == 1
    assert paragraphs[0].body == "\nテスト\n1\n"
    assert paragraphs[0].as_original_pos(6) == (2, 1)
    assert paragraphs[0].as_original_index(6) == 11



from japick.parser import html_parser


def test_h1():
    tags = html_parser("<h1>これはテストです</h1>")
    assert tags == [
        {
            "tag": "h1",
            "start": 0,
            "end": 17,
            "body": "<h1>これはテストです</h1>",
            "pos": (0, 0),
            "pos_end": (0, 17),
        },
    ]


def test_li_with_p():
    tags = html_parser("<ul><li><p>リスト</p></li></ul>")
    assert tags == [
        {
            "tag": "li",
            "start": 4,
            "end": 23,
            "body": "<li><p>リスト</p></li>",
            "pos": (0, 4),
            "pos_end": (0, 23),
        },
    ]


def test_tag_close_tag_text():
    tags = html_parser('<h1 data-example=">">これはテストです</h1>')
    assert tags == [
        {
            "tag": "h1",
            "start": 0,
            "end": 34,
            "body": '<h1 data-example=">">これはテストです</h1>',
            "pos": (0, 0),
            "pos_end": (0, 34),
        },
    ]


def test_with_spaces():
    tags = html_parser(
        """
    <section>
      これがセクションです。
    </section>
"""
    )
    assert tags == [
        {
            "tag": "section",
            "start": 5,
            "end": 47,
            "body": "<section>\n      これがセクションです。\n    </section>",
            "pos": (1, 4),
            "pos_end": (3, 14),
        }
    ]


def test_nested():
    tags = html_parser(
        """
    <section>
      これがセクションです。
      <div>
          子供です。
          <p>これが段落です。</p>
          子供の後ろです。
      </div>
    </section>
"""
    )
    assert tags == [
        {
            "tag": "section",
            "start": 5,
            "end": 39,
            "body": "<section>\n      これがセクションです。\n      ",
            "pos": (1, 4),
            "pos_end": (3, 6),
        },
        {
            "tag": "div",
            "start": 39,
            "end": 71,
            "body": "<div>\n          子供です。\n          ",
            "pos": (3, 6),
            "pos_end": (5, 10),
        },
        {
            "tag": "p",
            "start": 71,
            "end": 86,
            "body": "<p>これが段落です。</p>",
            "pos": (5, 10),
            "pos_end": (5, 25),
        },
        {
            "tag": "div",
            "start": 86,
            "end": 118,
            "body": "\n          子供の後ろです。\n      </div>",
            "pos": (5, 25),
            "pos_end": (7, 12),
        },
    ]


def test_nested_brother():
    tags = html_parser(
        """
    <section>
      <h1>見出しです</h1>
      <p>これが段落です。</p>
    </section>
"""
    )
    assert tags == [
        {
            "tag": "h1",
            "start": 21,
            "end": 35,
            "body": "<h1>見出しです</h1>",
            "pos": (2, 6),
            "pos_end": (2, 20),
        },
        {
            "tag": "p",
            "start": 42,
            "end": 57,
            "body": "<p>これが段落です。</p>",
            "pos": (3, 6),
            "pos_end": (3, 21),
        },
    ]


def test_self_closing_tag():
    tags = html_parser("<div/>")
    assert tags == []


def test_naked_text():
    tags = html_parser("これはテストです")
    assert tags == []


def test_windows_return():
    tags = html_parser("<p>テスト1</p>\r\n<p>テスト2</p>")
    assert tags == [
        {
            "tag": "p",
            "start": 0,
            "end": 11,
            "body": "<p>テスト1</p>",
            "pos": (0, 0),
            "pos_end": (0, 11),
        },
        {
            "tag": "p",
            "start": 13,
            "end": 24,
            "body": "<p>テスト2</p>",
            "pos": (1, 0),
            "pos_end": (1, 11),
        },
    ]



from html.parser import HTMLParser

from .syntax import HTML_REGEX

# https://developer.mozilla.org/ja/docs/Web/HTML/Element
TARGET_TAGS = [
    "title",
    "article",
    "header",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "main",
    "section",
    "search",
    "blockquote",
    "dd",
    "div",
    "dt",
    "figcaption",
    "figure",
    # "li",
    "p",
    "pre",
    "caption",
    "td",
    "th",
    "button",
    "label",
    "textarea",
    "details",
    "dialog",
    "summary",
]


class IndexHTMLParser(HTMLParser):
    """HTMLからTARGET_TAGSに指定された要素を抜き出し、その位置情報を保持する

    結果は入れ子構造を持たず、親は子供のテキストもすべて含む形で返し、子要素も同様に結果に含まれる
    ただしliタグは例外で、liタグ内の要素はliタグの範囲のみ処理される（子要素は含まれない）
    """

    def __init__(self, raw_html):
        super().__init__()
        self.raw_html = raw_html
        self.tags = []
        self.current_tags = []
        self.in_li = False

    def handle_starttag(self, tag, attrs):
        if tag == "li":
            # liタグ内の場合、他の要素を無視しliタグの範囲で処理
            # （<li><p>...</p></li> などがあり得るため）
            self.current_tags.append((tag, self.getpos_zero()))
            self.in_li = True
        elif tag in TARGET_TAGS and not self.in_li:
            self.current_tags.append((tag, self.getpos_zero()))

    def append_tag(self, tag=None):
        if not self.current_tags:
            return

        current_tag, pos = self.current_tags.pop()
        start = self.pos_to_index(pos)
        end = self.pos_to_index(self.getpos_zero())

        # 閉じタグの終了位置を取得する
        m = HTML_REGEX.search(self.raw_html[end:])
        if m:
            end += m.end()

        self.tags.append(
            {
                "tag": tag or current_tag,
                "start": start,
                "end": end,
                "body": self.raw_html[start:end],
                "pos": pos,
                "pos_end": self.index_to_pos(end),
            }
        )

    def handle_endtag(self, tag):
        if tag == "li" and self.in_li:
            self.append_tag("li")
            self.in_li = False
        elif self.current_tags and tag == self.current_tags[-1][0] and not self.in_li:
            self.append_tag()

    def getpos_zero(self):
        line, ch = self.getpos()
        return line - 1, ch

    def pos_to_index(self, pos):
        # (Line, Column) -> Index
        lines = self.raw_html.splitlines(keepends=True)
        return sum(len(line) for line in lines[: pos[0]]) + pos[1]

    def index_to_pos(self, index):
        lines = self.raw_html.splitlines(keepends=True)
        char_count = 0
        for line_num, line in enumerate(lines):
            line_length = len(line)
            if char_count + line_length > index:
                column = index - char_count
                return line_num, column
            char_count += line_length
        return len(lines) - 1, len(lines[-1].rstrip("\r\n"))


def split_nested_elements(elements):
    """入れ子構造を持つ要素を分割して返す関数

    たとえば以下のようなタグ要素があるとする：

        [
            {'start': 0, 'end': 23, 'body': '<div>A<p>Text</p></div>'},
            {'start': 6, 'end': 17, 'body': '<p>Text</p>'},
        ]

    この場合、1要素目のbodyに2要素目の結果がそのまま含まれてしまう。
    この関数はこのような入れ子構造を持つ要素を分割して返す：

        [
            {'start': 0, 'end': 6, 'body': '<div>A'},
            {'start': 6, 'end': 17, 'body': '<p>Text</p>'},
        ]

    ただし結果が空文字かHTML要素しか含まない場合は除外される
    """

    def split_parent(parent, child):
        # 親要素を子要素の前後で2つに分割する
        before_child = {
            **parent,
            "start": parent["start"],
            "end": child["start"],
            "pos": parent["pos"],
            "pos_end": child["pos"],
            "body": parent["body"][: child["start"] - parent["start"]],
        }
        after_child = {
            **parent,
            "start": child["end"],
            "end": parent["end"],
            "pos": child["pos_end"],
            "pos_end": parent["pos_end"],
            "body": parent["body"][child["end"] - parent["start"] :],
        }
        return before_child, after_child

    elements = sorted(elements, key=lambda x: (x["start"], x["end"]))
    new_elements = {}
    i = 0
    # 要素のすべてをループしつつ、自身の以降の要素と比較し入れ子かどうかを確認
    # 入れ子がある場合は、親要素を分割して新しい要素を作成し、分割した後半は再度処理する
    while i < len(elements):
        parent = elements[i]
        nested_found = False
        # 子要素の探索
        for j in range(i + 1, len(elements)):
            child = elements[j]
            if parent["start"] < child["start"] and child["end"] < parent["end"]:
                nested_found = True
                # 親要素を分割し、前半と子供を返す一覧に追加
                before_child, after_child = split_parent(parent, child)
                new_elements[(before_child["start"], before_child["end"])] = before_child
                # 親要素の後半をリストに追加してもう一回処理
                elements[i] = after_child
                break
        if not nested_found:
            new_elements[(parent["start"], parent["end"])] = parent
            i += 1  # 入れ子がない場合のみインクリメント

    new_elements = [el for el in new_elements.values() if HTML_REGEX.sub("", el["body"]).strip()]
    return sorted(new_elements, key=lambda x: (x["start"], x["end"]))


def html_parser(html: str):
    parser = IndexHTMLParser(html)
    parser.feed(html)
    return split_nested_elements(parser.tags)



from setuptools import setup, find_packages


setup(
    name="japick",
    version="0.6.2",
    packages=find_packages(),
    url="https://shodo.ink/",
    author="ZenProducts Inc.",
    author_email="info@shodo.ink",
    extras_require={
        "tests": ["pytest", "invoke", "black", "isort"],
    },
)
